@@ -1,6 +1,6 @@
 /* -*-pgsql-c-*- */
 /*
- * $Header: /home/t-ishii/repository/pgpool/child.c,v 1.24 2004/05/08 02:03:27 t-ishii Exp $
+ * $Header: /home/t-ishii/repository/pgpool/child.c,v 1.28 2004/06/15 12:57:55 t-ishii Exp $
  *
  * pgpool: a language independent connection pool server for PostgreSQL 
  * written by Tatsuo Ishii
@@ -37,6 +37,7 @@
 #include <errno.h>
 #include <string.h>
 #include <unistd.h>
+#include <stdlib.h>
 
 #ifdef HAVE_FCNTL_H
 #include <fcntl.h>
@@ -51,11 +52,11 @@ static void unset_nonblock(int fd);
 
 static POOL_CONNECTION *do_accept(int unix_fd, int inet_fd);
 static StartupPacket *read_startup_packet(POOL_CONNECTION *cp);
-static int send_startup_packet(POOL_CONNECTION *cp, StartupPacket *sp);
+static int send_startup_packet(POOL_CONNECTION_POOL_SLOT *cp);
 static POOL_CONNECTION_POOL *connect_backend(StartupPacket *sp, POOL_CONNECTION *frontend);
-static void reset_backend(POOL_CONNECTION *backend);
-static void cancel_request(CancelPacket *cp, int secondary_backend);
+static void cancel_request(CancelPacket *sp, int secondary_backend);
 static RETSIGTYPE die(int sig);
+static int send_params(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *backend);
 
 /*
 * child main loop
@@ -65,13 +66,9 @@ void do_child(int unix_fd, int inet_fd)
 	StartupPacket *sp;
 	POOL_CONNECTION *frontend;
 	POOL_CONNECTION_POOL *backend;
-#ifdef READ_PERFORMANCE
-	struct timeval now1, now2;
-	static long atime;
-	static int cnt;
-#endif
 
 	pool_debug("I am %d", getpid());
+
 	/* set up signal handlers */
 	signal(SIGALRM, SIG_DFL);
 	signal(SIGTERM, die);
@@ -97,7 +94,6 @@ void do_child(int unix_fd, int inet_fd)
 
 	for (;;)
 	{
-		int major;
 		int connection_reuse = 1;
 		int ssl_request = 0;
 
@@ -116,10 +112,16 @@ void do_child(int unix_fd, int inet_fd)
 		pool_disable_timeout();
 
 		/* read the startup packet */
+		sp = 0;
 	retry_startup:
-#ifdef READ_PERFORMANCE
-	gettimeofday(&now1,0);
-#endif
+		if (sp)
+		{
+			free(sp->startup_packet);
+			free(sp->database);
+			free(sp->user);
+			free(sp);
+		}
+
 		sp = read_startup_packet(frontend);
 		if (sp == NULL)
 		{
@@ -128,46 +130,19 @@ void do_child(int unix_fd, int inet_fd)
 			pool_close(frontend);
 			continue;
 		}
-#ifdef READ_PERFORMANCE
-	gettimeofday(&now2,0);
-	atime += (now2.tv_sec - now1.tv_sec)*1000000 + (now2.tv_usec - now1.tv_usec);
-	cnt++;
-	if (cnt % 100 == 0)
-	{
-		pool_log("cnt: %d atime: %ld", cnt, atime);
-	}
-#endif
-		major = ntohl(sp->protoVersion) >> 16;
-		pool_debug("major: %d", major);
 
-		/* V3(7.4 or later) protocol? */
-		if (major < 1234 && major > 2)
+		/* cancel request? */
+		if (sp->major == 1234 && sp->minor == 5678)
 		{
-			static char *msg = "Sorry, we do not support V3 protocol";
-
-			pool_debug("V3 protocol request received. attempt to fallback to V2");
-
-			/*
-			 * Send "we do not suppport V3 protocol" error in the hope that
-			 * it could fallback to V2 protocol..,.
-			 */
-			pool_write_and_flush(frontend, "E", 1);
-			pool_write_and_flush(frontend, msg, strlen(msg)+1);
+			cancel_request((CancelPacket *)sp->startup_packet, 0);
+			if (pool_config.replication_enabled)
+				cancel_request((CancelPacket *)sp->startup_packet, 1);
 			pool_close(frontend);
 			continue;
 		}
 
-		/* cancel request? */
-		if (ntohl(sp->protoVersion) == 0x04D2162E)		/* 0x04D2162F = 12345678 */
-		{
-			cancel_request((CancelPacket *)sp, 0);
-			if (pool_config.replication_enabled)
-				cancel_request((CancelPacket *)sp, 1);
-			continue;
-		}
-
 		/* SSL? */
-		if (ntohl(sp->protoVersion) == 0x04D2162F)		/* 0x04D2162F = 12345679 */
+		if (sp->major == 1234 && sp->minor == 5679)
 		{
 			/* SSL not supported */
 			pool_debug("SSLRequest: sent N; retry startup");
@@ -186,12 +161,17 @@ void do_child(int unix_fd, int inet_fd)
 			goto retry_startup;
 		}
 
+		/*
+		 * Ok, negotiaton with frontend has been done. Let's go to the next step.
+		 */
+
 		/*
 		 * if there's no connection associated with user and database,
 		 * we need to connect to the backend and send the startup packet.
 		 */
-		if ((backend = pool_get_cp(sp->user, sp->database)) == NULL)
+		if ((backend = pool_get_cp(sp->user, sp->database, sp->major)) == NULL)
 		{
+			/* create a new connection to backend */
 			connection_reuse = 0;
 
 			if ((backend = connect_backend(sp, frontend)) == NULL)
@@ -199,20 +179,43 @@ void do_child(int unix_fd, int inet_fd)
 		}
 		else
 		{
-			reset_backend(MASTER_CONNECTION(backend)->con);
-			if (pool_config.replication_enabled)
-				reset_backend(SECONDARY_CONNECTION(backend)->con);
+			/* reuse existing connection to backend */
 
 			if (pool_do_reauth(frontend, backend))
 			{
 				pool_close(frontend);
 				continue;
 			}
-			if (pool_write_and_flush(frontend, "Z", 1) < 0)
+
+			if (MAJOR(backend) == 3)
+			{
+				if (send_params(frontend, backend))
+				{
+					pool_close(frontend);
+					continue;
+				}
+			}
+
+			/* send ReadyForQuery to frontend */
+			pool_write(frontend, "Z", 1);
+
+			if (MAJOR(backend) == 3)
+			{
+				int len;
+				char tstate;
+
+				len = htonl(5);
+				pool_write(frontend, &len, sizeof(len));
+				tstate = TSTATE(backend);
+				pool_write(frontend, &tstate, 1);
+			}
+
+			if (pool_flush(frontend) < 0)
 			{
 				pool_close(frontend);
 				continue;
 			}
+
 		}
 
 		/* enable query result read timeout if non strict mode */
@@ -222,24 +225,28 @@ void do_child(int unix_fd, int inet_fd)
 		/* query process loop */
 		for (;;)
 		{
-			int status;
+			POOL_STATUS status;
 
-			status = pool_process_query(frontend, backend, connection_reuse);
+			status = pool_process_query(frontend, backend, 0);
 			switch (status)
 			{
 				/* client exits */
 				case POOL_END:
-					pool_close(frontend);
-
 					/* do not cache connection to template0, template1, regression */
 					if (!strcmp(sp->database, "template0") || !strcmp(sp->database, "template1") ||
 						!strcmp(sp->database, "regression"))
 					{
+						pool_close(frontend);
 						pool_send_frontend_exits(backend);
-						pool_discard_cp(sp->user, sp->database);
+						pool_discard_cp(sp->user, sp->database, sp->major);
 					}
 					else
+					{
+						/* send reset request to backend */
+						pool_process_query(frontend, backend, 1);
+						pool_close(frontend);
 						pool_connection_pool_timer(backend);
+					}
 					break;
 				
 				/* error occured. discard backend connection pool
@@ -248,8 +255,8 @@ void do_child(int unix_fd, int inet_fd)
 #ifdef NOT_USED
 					pool_discard_cp(sp->user, sp->database);
 					pool_close(frontend);
-#endif
 					notice_backend_error();
+#endif
 					exit(1);
 					break;
 
@@ -434,11 +441,20 @@ static POOL_CONNECTION *do_accept(int unix_fd, int inet_fd)
 */
 static StartupPacket *read_startup_packet(POOL_CONNECTION *cp)
 {
-	static StartupPacket sp;
+	StartupPacket *sp;
+	StartupPacket_v2 *sp2;
+	int protov;
 	int len;
+	char *p;
 
-	memset((char *)&sp, 0, sizeof(sp));
+	sp = (StartupPacket *)malloc(sizeof(StartupPacket));
+	if (!sp)
+	{
+		pool_error("read_startup_packet: out of memory");
+		return NULL;
+	}
 
+	/* read startup packet length */
 	if (pool_read(cp, &len, sizeof(len)))
 	{
 		return NULL;
@@ -446,83 +462,125 @@ static StartupPacket *read_startup_packet(POOL_CONNECTION *cp)
 	len = ntohl(len);
 	len -= sizeof(len);
 
-	if (pool_read(cp, &sp, len))
+	if (len <= 0)
+	{
+		pool_error("read_startup_packet: incorrect packet length (%d)", len);
+	}
+
+	sp->startup_packet = calloc(len, 1);
+	if (!sp->startup_packet)
+	{
+		pool_error("read_startup_packet: out of memory");
+		return NULL;
+	}
+
+	/* read startup packet */
+	if (pool_read(cp, sp->startup_packet, len))
 	{
 		return NULL;
 	}
 
-	pool_debug("Protocol Version: %08x",ntohl(sp.protoVersion));
-	pool_debug("Protocol Major: %d Minor:%d",
-		   ntohl(sp.protoVersion)>>16,
-			   ntohl(sp.protoVersion) & 0x0000ffff);
-	pool_debug("database: %s",sp.database);
-	pool_debug("user: %s",sp.user);
+	sp->len = len;
+	memcpy(&protov, sp->startup_packet, sizeof(protov));
+	sp->major = ntohl(protov)>>16;
+	sp->minor = ntohl(protov) & 0x0000ffff;
+	p = sp->startup_packet;
 
-	return &sp;
+	switch(sp->major)
+	{
+		case PROTO_MAJOR_V2: /* V2 */
+			sp2 = (StartupPacket_v2 *)(sp->startup_packet);
+
+			sp->database = calloc(SM_DATABASE+1, 1);
+			if (!sp->database)
+			{
+				pool_error("read_startup_packet: out of memory");
+				return NULL;
+			}
+			strncpy(sp->database, sp2->database, SM_DATABASE);
+
+			sp->user = calloc(SM_USER+1, 1);
+			if (!sp->user)
+			{
+				pool_error("read_startup_packet: out of memory");
+				return NULL;
+			}
+			strncpy(sp->user, sp2->user, SM_USER);
+
+			break;
+
+		case PROTO_MAJOR_V3: /* V3 */
+			p += sizeof(int);	/* skip protocol version info */
+
+			while(*p)
+			{
+				if (!strcmp("user", p))
+				{
+					p += (strlen(p) + 1);
+					sp->user = strdup(p);
+					if (!sp->user)
+					{
+						pool_error("read_startup_packet: out of memory");
+						return NULL;
+					}
+				}
+				else if (!strcmp("database", p))
+				{
+					p += (strlen(p) + 1);
+					sp->database = strdup(p);
+					if (!sp->database)
+					{
+						pool_error("read_startup_packet: out of memory");
+						return NULL;
+					}
+				}
+				p += (strlen(p) + 1);
+			}
+			break;
+
+		case 1234:		/* cancel or SSL request */
+			/* set dummy database, user info */
+			sp->database = calloc(1, 1);
+			if (!sp->database)
+			{
+				pool_error("read_startup_packet: out of memory");
+				return NULL;
+			}
+			sp->user = calloc(1, 1);
+			if (!sp->user)
+			{
+				pool_error("read_startup_packet: out of memory");
+				return NULL;
+			}
+			break;
+
+		default:
+			pool_error("read_startup_packet: invalid major no: %d", sp->major);
+			return NULL;
+	}
+
+	pool_debug("Protocol Major: %d Minor: %d database: %s user: %s", 
+			   sp->major, sp->minor, sp->database, sp->user);
+
+	return sp;
 }
 
 /*
 * send startup packet
 */
-static int send_startup_packet(POOL_CONNECTION *cp, StartupPacket *sp)
+static int send_startup_packet(POOL_CONNECTION_POOL_SLOT *cp)
 {
 	int len;
 
-	len = htonl(sizeof(len) + sizeof(StartupPacket));
-
-	pool_write(cp, &len, sizeof(len));
-	return pool_write_and_flush(cp, sp, sizeof(StartupPacket));
+	len = htonl(cp->sp->len + sizeof(len));
+	pool_write(cp->con, &len, sizeof(len)); 
+	return pool_write_and_flush(cp->con, cp->sp->startup_packet, cp->sp->len);
 }
 
 /*
- * reset backend status
+ * process cancel request
  */
-static void reset_backend(POOL_CONNECTION *backend)
-{
-	char kind;
-	int len;
-
-#ifdef NO_RESET_ALL
-	static char *queries[] = {"ABORT"};
-#else
-	static char *queries[] = {"ABORT", "RESET ALL"};
-#endif
-
-	int i;
-
-	for (i=0;i<sizeof(queries)/sizeof(char *);i++)
-	{
-		pool_write(backend, "Q", 1);
-		pool_write_and_flush(backend, queries[i], strlen(queries[i])+1);
-		pool_read(backend, &kind, 1);
-
-		switch(kind)
-		{
-			case 'C':
-			case 'E':
-			case 'N':
-				/* read and discard notice or error message */
-				pool_read_string(backend, &len, 0);
-
-				if (kind != 'C')
-				{
-					/* read and discard complete command response */
-					pool_read(backend, &kind, 1);
-					pool_read_string(backend, &len, 0);
-				}
-
-				/* read and discard ready for query message */
-				pool_read(backend, &kind, 1);
-				break;
-
-			default:
-				pool_error("reset_backend: Unknown response");
-				exit(1);
-		}
-	}
-}
-
-static void cancel_request(CancelPacket *cp, int secondary_backend)
+static void cancel_request(CancelPacket *sp, int secondary_backend)
 {
 	int	len;
 	int fd;
@@ -558,7 +616,7 @@ static void cancel_request(CancelPacket *cp, int secondary_backend)
 	len = htonl(sizeof(len) + sizeof(CancelPacket));
 	pool_write(con, &len, sizeof(len));
 
-	if (pool_write_and_flush(con, cp, sizeof(CancelPacket)) < 0)
+	if (pool_write_and_flush(con, sp, sizeof(CancelPacket)) < 0)
 		pool_error("Could not send cancel request packet");
 	pool_close(con);
 }
@@ -567,30 +625,35 @@ static POOL_CONNECTION_POOL *connect_backend(StartupPacket *sp, POOL_CONNECTION
 {
 	POOL_CONNECTION_POOL *backend;
 
-	/* connect to the backend and send the startup packect here */
-	backend = pool_create_cp(sp->user, sp->database);
+	/* connect to the backend */
+	backend = pool_create_cp();
 	if (backend == NULL)
 	{
-		static char *msg = "Sorry, too many clients already";
-
-		pool_error("do_child: cannot not create backend connection");
-
-		pool_write(frontend, "E", 1);
-		pool_write_and_flush(frontend, msg, strlen(msg)+1);
+		pool_send_error_message(frontend, sp->major, "XX000", "connection cache is full", "",
+								"increace max_pool", __FILE__, __LINE__);
 		pool_close(frontend);
 		return NULL;
 	}
 
 	/* mark this is a backend connection */
 	backend->slots[0]->con->isbackend = 1;
+	/*
+	 * save startup packet info
+	 */
+	backend->slots[0]->sp = sp;
+
 	if (pool_config.replication_enabled)
 	{
 		backend->slots[1]->con->isbackend = 1;
 		backend->slots[1]->con->issecondary_backend = 1;
+		/*
+		 * save startup packet info
+		 */
+		backend->slots[1]->sp = sp;
 	}
 
 	/* send startup packet */
-	if (send_startup_packet(backend->slots[0]->con, sp) < 0)
+	if (send_startup_packet(backend->slots[0]) < 0)
 	{
 		pool_error("do_child: fails to send startup packet to the backend");
 		pool_close(frontend);
@@ -600,7 +663,7 @@ static POOL_CONNECTION_POOL *connect_backend(StartupPacket *sp, POOL_CONNECTION
 	/* send startup packet */
 	if (pool_config.replication_enabled)
 	{
-		if (send_startup_packet(backend->slots[1]->con, sp) < 0)
+		if (send_startup_packet(backend->slots[1]) < 0)
 		{
 			pool_error("do_child: fails to send startup packet to the secondary backend");
 			pool_close(frontend);
@@ -614,7 +677,7 @@ static POOL_CONNECTION_POOL *connect_backend(StartupPacket *sp, POOL_CONNECTION
 	if (pool_do_auth(frontend, backend))
 	{
 		pool_close(frontend);
-		pool_discard_cp(sp->user, sp->database);
+		pool_discard_cp(sp->user, sp->database, sp->major);
 		return NULL;
 	}
 	return backend;
@@ -624,3 +687,28 @@ static RETSIGTYPE die(int sig)
 {
 	exit(0);
 }
+
+static int send_params(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *backend)
+{
+	int index;
+	char *name, *value;
+	int len, sendlen;
+
+	index = 0;
+	while (pool_get_param(&MASTER(backend)->params, index++, &name, &value) == 0)
+	{
+		pool_write(frontend, "S", 1);
+		len = sizeof(sendlen) + strlen(name) + 1 + strlen(value) + 1;
+		sendlen = htonl(len);
+		pool_write(frontend, &sendlen, sizeof(sendlen));
+		pool_write(frontend, name, strlen(name) + 1);
+		pool_write(frontend, value, strlen(value) + 1);
+	}
+
+	if (pool_flush(frontend))
+	{
+		pool_error("pool_send_params: pool_flush() failed");
+		return -1;
+	}
+	return 0;
+}
