@@ -1,6 +1,6 @@
 /* -*-pgsql-c-*- */
 /*
- * $Header: /home/t-ishii/repository/pgpool/pool_connection_pool.c,v 1.17 2004/05/06 06:37:02 t-ishii Exp $
+ * $Header: /home/t-ishii/repository/pgpool/pool_connection_pool.c,v 1.20 2004/06/10 10:25:22 t-ishii Exp $
  *
  * pgpool: a language independent connection pool server for PostgreSQL 
  * written by Tatsuo Ishii
@@ -36,16 +36,16 @@
 #include <signal.h>
 #include <string.h>
 #include <unistd.h>
+#include <stdlib.h>
 
 #include "pool.h"
 
 POOL_CONNECTION_POOL *pool_connection_pool;	/* connection pool */
 
-static POOL_CONNECTION_POOL_SLOT *create_cp(POOL_CONNECTION_POOL_SLOT *cp,
-											char *user, char *database,
-											int secondary_backend);
+static POOL_CONNECTION_POOL_SLOT *create_cp(POOL_CONNECTION_POOL_SLOT *cp, int secondary_backend);
+
 /*
-* initialize connection pool
+* initialize connection pools. this should be called once at the startup.
 */
 int pool_init_cp(void)
 {
@@ -63,7 +63,7 @@ int pool_init_cp(void)
 /*
 * find connection by user and database
 */
-POOL_CONNECTION_POOL *pool_get_cp(char *user, char *database)
+POOL_CONNECTION_POOL *pool_get_cp(char *user, char *database, int protoMajor)
 {
 	int i;
 
@@ -77,9 +77,11 @@ POOL_CONNECTION_POOL *pool_get_cp(char *user, char *database)
 
 	for (i=0;i<pool_config.max_pool;i++)
 	{
-		if (MASTER_CONNECTION(p) && MASTER_CONNECTION(p)->user != NULL &&
-			strcmp(MASTER_CONNECTION(p)->user, user) == 0 &&
-			strcmp(MASTER_CONNECTION(p)->database, database) == 0)
+		if (MASTER_CONNECTION(p) &&
+			MASTER_CONNECTION(p)->sp->major == protoMajor &&
+			MASTER_CONNECTION(p)->sp->user != NULL &&
+			strcmp(MASTER_CONNECTION(p)->sp->user, user) == 0 &&
+			strcmp(MASTER_CONNECTION(p)->sp->database, database) == 0)
 		{
 			/* mark this connection is under use */
 			MASTER_CONNECTION(p)->closetime = 0;
@@ -93,11 +95,11 @@ POOL_CONNECTION_POOL *pool_get_cp(char *user, char *database)
 /*
  * disconnect and release a connection to the database
  */
-void pool_discard_cp(char *user, char *database)
+void pool_discard_cp(char *user, char *database, int protoMajor)
 {
 	int i;
 
-	POOL_CONNECTION_POOL *p = pool_get_cp(user, database);
+	POOL_CONNECTION_POOL *p = pool_get_cp(user, database, protoMajor);
 
 	if (p == NULL)
 	{
@@ -105,14 +107,14 @@ void pool_discard_cp(char *user, char *database)
 		return;
 	}
 
-	free(MASTER_CONNECTION(p)->user);
-	free(MASTER_CONNECTION(p)->database);
+	free(MASTER_CONNECTION(p)->sp->user);
+	free(MASTER_CONNECTION(p)->sp->database);
+	free(MASTER_CONNECTION(p)->sp->startup_packet);
 	pool_close(MASTER_CONNECTION(p)->con);
 
 	if (pool_config.replication_enabled)
 	{
-		free(SECONDARY_CONNECTION(p)->user);
-		free(SECONDARY_CONNECTION(p)->database);
+		/* do not free memory! we did not allocate them */
 		pool_close(SECONDARY_CONNECTION(p)->con);
 	}
 
@@ -121,9 +123,9 @@ void pool_discard_cp(char *user, char *database)
 
 
 /*
-* create connection pool by user and database
+* create a connection pool by user and database
 */
-POOL_CONNECTION_POOL *pool_create_cp(char *user, char *database)
+POOL_CONNECTION_POOL *pool_create_cp(void)
 {
 	int i;
 	int fd;
@@ -142,15 +144,38 @@ POOL_CONNECTION_POOL *pool_create_cp(char *user, char *database)
 		{
 			/* create master connection */
 			MASTER_CONNECTION(p) = malloc(sizeof(POOL_CONNECTION_POOL_SLOT));
-			MASTER_CONNECTION(p) = create_cp(MASTER_CONNECTION(p), user, database, 0);
-			p->num = 1;
+			if (MASTER_CONNECTION(p) == NULL)
+			{
+				pool_error("pool_create_cp: malloc() failed");
+				return NULL;
+			}
+			create_cp(MASTER_CONNECTION(p), 0);
+
+			/* initialize Paramter Status save structure */
+			if (pool_init_params(&MASTER(p)->params))
+			{
+				return NULL;
+			}
+			p->num = 1;	/* number of slots */
 
 			/* create secondary connection */
 			if (pool_config.replication_enabled)
 			{
 				SECONDARY_CONNECTION(p) = malloc(sizeof(POOL_CONNECTION_POOL_SLOT));
-				SECONDARY_CONNECTION(p) = create_cp(SECONDARY_CONNECTION(p), user, database, 1);
-				p->num++;
+				if (SECONDARY_CONNECTION(p) == NULL)
+				{
+					pool_error("pool_create_cp: malloc() failed");
+					return NULL;
+				}
+				create_cp(SECONDARY_CONNECTION(p), 1);
+
+				/* initialize Paramter Status save structure */
+				if (pool_init_params(&MASTER(p)->params))
+				{
+					return NULL;
+				}
+
+				p->num++;	/* number of slots */
 			}
 
 			return p;
@@ -180,7 +205,7 @@ void pool_connection_pool_timer(POOL_CONNECTION_POOL *backend)
 	{
 		if (!MASTER_CONNECTION(p))
 			continue;
-		if (MASTER_CONNECTION(p)->user == NULL)
+		if (MASTER_CONNECTION(p)->sp->user == NULL)
 			continue;
 
 		if (p != backend && MASTER_CONNECTION(p)->closetime)
@@ -213,7 +238,7 @@ RETSIGTYPE pool_backend_timer_handler(int sig)
 	{
 		if (!MASTER_CONNECTION(p))
 			continue;
-		if (MASTER_CONNECTION(p)->user == NULL)
+		if (MASTER_CONNECTION(p)->sp->user == NULL)
 			continue;
 
 		/* timer expire? */
@@ -225,19 +250,21 @@ RETSIGTYPE pool_backend_timer_handler(int sig)
 			if (now >= (MASTER_CONNECTION(p)->closetime+pool_config.connection_life_time))
 			{
 				/* discard expired connection */
-				pool_debug("pool_backend_timer_handler: expires user %s database %s", MASTER_CONNECTION(p)->user, MASTER_CONNECTION(p)->database);
+				pool_debug("pool_backend_timer_handler: expires user %s database %s", MASTER_CONNECTION(p)->sp->user, MASTER_CONNECTION(p)->sp->database);
 
 				pool_send_frontend_exits(p);
 
-				free(MASTER_CONNECTION(p)->user);
-				free(MASTER_CONNECTION(p)->database);
+				free(MASTER_CONNECTION(p)->sp->user);
+				free(MASTER_CONNECTION(p)->sp->database);
+				free(MASTER_CONNECTION(p)->sp->startup_packet);
 				pool_close(MASTER_CONNECTION(p)->con);
 				memset(p, 0, sizeof(POOL_CONNECTION_POOL));
 
 				if (pool_config.replication_enabled)
 				{
-					free(SECONDARY_CONNECTION(p)->user);
-					free(SECONDARY_CONNECTION(p)->database);
+					free(SECONDARY_CONNECTION(p)->sp->user);
+					free(SECONDARY_CONNECTION(p)->sp->database);
+					free(SECONDARY_CONNECTION(p)->sp->startup_packet);
 					pool_close(SECONDARY_CONNECTION(p)->con);
 					memset(p, 0, sizeof(POOL_CONNECTION_POOL));
 				}
@@ -352,8 +379,7 @@ int connect_unix_domain_socket(int secondary_backend)
 	return fd;
 }
 
-static POOL_CONNECTION_POOL_SLOT *create_cp(POOL_CONNECTION_POOL_SLOT *cp, char *user, char *database,
-									   int secondary_backend)
+static POOL_CONNECTION_POOL_SLOT *create_cp(POOL_CONNECTION_POOL_SLOT *cp, int secondary_backend)
 {
 	int fd;
 
@@ -369,19 +395,6 @@ static POOL_CONNECTION_POOL_SLOT *create_cp(POOL_CONNECTION_POOL_SLOT *cp, char
 		exit(1);
 	}
 
-	cp->user = strdup(user);
-	if (!cp->user)
-	{
-		pool_error("pool_create_cp: out of memory");
-		return NULL;
-	}
-
-	cp->database = strdup(database);
-	if (!cp->database)
-	{
-		pool_error("pool_create_cp: out of memory");
-		return NULL;
-	}
 	cp->con = pool_open(fd);
 	cp->closetime = 0;
 	return cp;
