@@ -1,6 +1,6 @@
 /* -*-pgsql-c-*- */
 /*
- * $Header: /home/t-ishii/repository/pgpool/pool_process_query.c,v 1.16.2.1 2004/05/16 03:27:32 t-ishii Exp $
+ * $Header: /home/t-ishii/repository/pgpool/pool_process_query.c,v 1.22 2004/06/18 12:11:06 t-ishii Exp $
  *
  * pgpool: a language independent connection pool server for PostgreSQL 
  * written by Tatsuo Ishii
@@ -36,15 +36,11 @@
 
 #include "pool.h"
 
-#define REPLICATION (pool_config.replication_enabled)
-#define MASTER(p) MASTER_CONNECTION(p)->con
-#define SECONDARY(p) SECONDARY_CONNECTION(p)->con
-
 static POOL_STATUS NotificationResponse(POOL_CONNECTION *frontend, 
 										POOL_CONNECTION_POOL *backend);
 
 static POOL_STATUS Query(POOL_CONNECTION *frontend, 
-						 POOL_CONNECTION_POOL *backend);
+						 POOL_CONNECTION_POOL *backend, char *query);
 
 static POOL_STATUS ReadyForQuery(POOL_CONNECTION *frontend, 
 								 POOL_CONNECTION_POOL *backend, int send_ready);
@@ -64,11 +60,6 @@ static POOL_STATUS CopyDataRows(POOL_CONNECTION *frontend,
 static POOL_STATUS CursorResponse(POOL_CONNECTION *frontend, 
 								  POOL_CONNECTION_POOL *backend);
 
-#ifdef NOTUSED
-static POOL_STATUS CancelRequest(POOL_CONNECTION *frontend, 
-								 POOL_CONNECTION_POOL *backend);
-#endif
-
 static POOL_STATUS NoticeResponse(POOL_CONNECTION *frontend, 
 								  POOL_CONNECTION_POOL *backend);
 
@@ -96,7 +87,15 @@ static POOL_STATUS ProcessFrontendResponse(POOL_CONNECTION *frontend,
 											POOL_CONNECTION_POOL *backend);
 
 static int synchronize(POOL_CONNECTION *cp);
-static void process_reporting(POOL_CONNECTION *frontend);
+static void process_reporting(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *backend);
+static int reset_backend(POOL_CONNECTION_POOL *backend, int qcnt);
+
+static int load_balance_enabled(POOL_CONNECTION_POOL *backend, char *sql);
+static void start_load_balance(POOL_CONNECTION_POOL *backend);
+static void end_load_balance(POOL_CONNECTION_POOL *backend);
+
+static POOL_CONNECTION_POOL_SLOT *slots[MAX_CONNECTION_SLOTS];
+static int in_load_balance;
 
 POOL_STATUS pool_process_query(POOL_CONNECTION *frontend, 
 							   POOL_CONNECTION_POOL *backend,
@@ -110,21 +109,49 @@ POOL_STATUS pool_process_query(POOL_CONNECTION *frontend,
 	fd_set	exceptmask;
 	int fds;
 	POOL_STATUS status;
+	int state;	/* 0: ok to issue commands 1: waiting for "ready for query" response */
+	int qcnt;
 
-	pool_debug("pool_process_query");
-
-	if (connection_reuse)
-	{
-		status = ReadyForQuery(frontend, backend, 0);
-		if (status != POOL_CONTINUE)
-			return status;
-	}
+	frontend->no_forward = connection_reuse;
+	qcnt = 0;
+	state = 0;
 
 	for (;;)
 	{
 		kind = kind1 = 0;
 		fkind = 0;
 
+		if (state == 0 && connection_reuse)
+		{
+			int st;
+
+			/* send query for resetting connection such as "ROLLBACK" "RESET ALL"... */
+			st = reset_backend(backend, qcnt);
+
+			if (st < 0)		/* error? */
+				return POOL_END;
+
+			else if (st == 0)	/* no query issued? */
+			{
+				qcnt++;
+				continue;
+			}
+
+			else if (st == 1)	/* more query remains */
+			{
+				state = 1;
+				qcnt++;
+				continue;
+			}
+
+			else if (st == 2)	/* no more qury */
+			{
+				frontend->no_forward = 0;
+				return POOL_CONTINUE;
+			}
+
+		}
+
 		if (!REPLICATION && MASTER(backend)->len == 0 && frontend->len == 0 ||
 			REPLICATION && MASTER(backend)->len == 0 &&
 			SECONDARY(backend)->len == 0
@@ -139,23 +166,33 @@ POOL_STATUS pool_process_query(POOL_CONNECTION *frontend,
 			FD_ZERO(&readmask);
 			FD_ZERO(&writemask);
 			FD_ZERO(&exceptmask);
-			FD_SET(frontend->fd, &readmask);
+			if (!connection_reuse)
+				FD_SET(frontend->fd, &readmask);
 			FD_SET(MASTER(backend)->fd, &readmask);
 			if (REPLICATION)
 				FD_SET(SECONDARY(backend)->fd, &readmask);
-			FD_SET(frontend->fd, &exceptmask);
+			if (!connection_reuse)
+				FD_SET(frontend->fd, &exceptmask);
 			FD_SET(MASTER(backend)->fd, &exceptmask);
 
-#ifdef DEBUG
-			pool_debug("pid: %d select in child", getpid());
-#endif
-			if (REPLICATION)
-				fds = select(Max(SECONDARY(backend)->fd,
-								 Max(frontend->fd, MASTER(backend)->fd)+1),
-							 &readmask, &writemask, &exceptmask, NULL);
+			if (connection_reuse)
+			{
+				if (REPLICATION)
+					fds = select(Max(SECONDARY(backend)->fd, MASTER(backend)->fd) + 1,
+								 &readmask, &writemask, &exceptmask, NULL);
+				else
+					fds = select(MASTER(backend)->fd+1, &readmask, &writemask, &exceptmask, NULL);
+			}
 			else
-				fds = select(Max(frontend->fd, MASTER(backend)->fd)+1,
-							 &readmask, &writemask, &exceptmask, NULL);
+			{
+				if (REPLICATION)
+					fds = select(Max(SECONDARY(backend)->fd,
+									 Max(frontend->fd, MASTER(backend)->fd)+1),
+								 &readmask, &writemask, &exceptmask, NULL);
+				else
+					fds = select(Max(frontend->fd, MASTER(backend)->fd)+1,
+								 &readmask, &writemask, &exceptmask, NULL);
+			}
 
 			if (fds == -1)
 			{
@@ -163,7 +200,7 @@ POOL_STATUS pool_process_query(POOL_CONNECTION *frontend,
 					continue;
 
 				pool_error("select() failed. reason: %s", strerror(errno));
-				return POOL_FATAL;
+				return POOL_ERROR;
 			}
 
 			if (fds == 0)
@@ -183,16 +220,16 @@ POOL_STATUS pool_process_query(POOL_CONNECTION *frontend,
 				pool_debug("read kind from secondary backend %c", kind1);
 			}
 
-			if (FD_ISSET(frontend->fd, &exceptmask))
+			if (!connection_reuse && FD_ISSET(frontend->fd, &exceptmask))
 			{
 				return POOL_END;
 			}
 			if (FD_ISSET(MASTER(backend)->fd, &exceptmask))
 			{
-				return POOL_FATAL;
+				return POOL_ERROR;
 			}
 
-			if (FD_ISSET(frontend->fd, &readmask))
+			if (!connection_reuse && FD_ISSET(frontend->fd, &readmask))
 			{
 				status = ProcessFrontendResponse(frontend, backend);
 				if (status != POOL_CONTINUE)
@@ -213,7 +250,15 @@ POOL_STATUS pool_process_query(POOL_CONNECTION *frontend,
 					{
 						pool_error("pool_process_query: kind does not match between backends master(%c) secondary(%c)",
 								   kind, kind1);
-						return POOL_FATAL;
+
+						pool_send_error_message(frontend, MAJOR(backend), "XX000", 
+												"kind mismatch between backends", "",
+												"check data consistency between master and secondary", __FILE__, __LINE__);
+
+						if (pool_config.replication_stop_on_mismatch)
+							return POOL_FATAL;
+						else
+							return POOL_ERROR;
 					}
 				}
 				pool_debug("read kind from backend pending data %c len: %d po: %d", kind, MASTER(backend)->len, MASTER(backend)->po);
@@ -243,116 +288,186 @@ POOL_STATUS pool_process_query(POOL_CONNECTION *frontend,
 			{
 				pool_error("pool_process_query: kind does not match between backends master(%c) secondary(%c)",
 						   kind, kind1);
-				return POOL_FATAL;
+				pool_send_error_message(frontend, MAJOR(backend), "XX000", 
+										"kind mismatch between backends", "",
+										"check data consistency between master and secondary", __FILE__, __LINE__);
+
+				if (pool_config.replication_stop_on_mismatch)
+					return POOL_FATAL;
+				else
+					return POOL_ERROR;
 			}
 		}
 
-		/* Prrocess backend Response */
-		switch (kind)
+		/*
+		 * Prrocess backend Response
+		 */
+
+		if (MAJOR(backend) == PROTO_MAJOR_V3)
 		{
-			case 'A':
-				/* Notification  response */
-				status = NotificationResponse(frontend, backend);
-				break;
-
-			case 'B':
-				/* BinaryRow */
-				status = BinaryRow(frontend, backend, num_fields);
-				break;
-
-			case 'C':
-				/* Complete command response */
-				status = CompleteCommandResponse(frontend, backend);
-				break;
-
-			case 'D':
-				/* AsciiRow */
-				status = AsciiRow(frontend, backend, num_fields);
-				break;
-
-			case 'E':
-				/* Error Response */
-				status = ErrorResponse(frontend, backend);
-				break;
-
-			case 'G':
-				/* CopyIn Response */
-				status = CopyInResponse(frontend, backend);
-				break;
-
-			case 'H':
-				/* CopyOut Response */
-				status = CopyOutResponse(frontend, backend);
-				break;
-
-			case 'I':
-				/* Empty Query Response */
-				status = EmptyQueryResponse(frontend, backend);
-				break;
-
-			case 'N':
-				/* Notice Response */
-				status = NoticeResponse(frontend, backend);
-				break;
-
-			case 'P':
-				/* CursorResponse */
-				status = CursorResponse(frontend, backend);
-				break;
-
-			case 'T':
-				/* RowDescription */
-				status = RowDescription(frontend, backend);
-				if (status < 0)
-					return POOL_ERROR;
-
-				num_fields = status;
-				status = POOL_CONTINUE;
-				break;
-
-			case 'V':
-				/* FunctionResultResponse and FunctionVoidResponse */
-				status = FunctionResultResponse(frontend, backend);
-				break;
-				
-			case 'Z':
-				/* Ready for query */
-				status = ReadyForQuery(frontend, backend, 1);
-				break;
-				
-			default:
-				pool_error("Unknown message type %c(%02x)", kind, kind);
-				exit(1);
+			switch (kind)
+			{
+				case 'G':
+					/* CopyIn response */
+					status = CopyInResponse(frontend, backend);
+					break;
+				case 'S':
+					/* Paramter Status */
+					status = ParameterStatus(frontend, backend);
+					break;
+				case 'Z':
+					/* Ready for query */
+					status = ReadyForQuery(frontend, backend, 1);
+					break;
+				default:
+					status = SimpleForwardToFrontend(kind, frontend, backend);
+					break;
+			}
 		}
+		else
+		{
+			switch (kind)
+			{
+				case 'A':
+					/* Notification  response */
+					status = NotificationResponse(frontend, backend);
+					break;
+
+				case 'B':
+					/* BinaryRow */
+					status = BinaryRow(frontend, backend, num_fields);
+					break;
+
+				case 'C':
+					/* Complete command response */
+					status = CompleteCommandResponse(frontend, backend);
+					break;
+
+				case 'D':
+					/* AsciiRow */
+					status = AsciiRow(frontend, backend, num_fields);
+					break;
+
+				case 'E':
+					/* Error Response */
+					status = ErrorResponse(frontend, backend);
+					break;
+
+				case 'G':
+					/* CopyIn Response */
+					status = CopyInResponse(frontend, backend);
+					break;
+
+				case 'H':
+					/* CopyOut Response */
+					status = CopyOutResponse(frontend, backend);
+					break;
+
+				case 'I':
+					/* Empty Query Response */
+					status = EmptyQueryResponse(frontend, backend);
+					break;
+
+				case 'N':
+					/* Notice Response */
+					status = NoticeResponse(frontend, backend);
+					break;
+
+				case 'P':
+					/* CursorResponse */
+					status = CursorResponse(frontend, backend);
+					break;
+
+				case 'T':
+					/* RowDescription */
+					status = RowDescription(frontend, backend);
+					if (status < 0)
+						return POOL_ERROR;
+
+					num_fields = status;
+					status = POOL_CONTINUE;
+					break;
+
+				case 'V':
+					/* FunctionResultResponse and FunctionVoidResponse */
+					status = FunctionResultResponse(frontend, backend);
+					break;
+				
+				case 'Z':
+					/* Ready for query */
+					status = ReadyForQuery(frontend, backend, 1);
+					break;
+				
+				default:
+					pool_error("Unknown message type %c(%02x)", kind, kind);
+					exit(1);
+			}
+		}
+
 		if (status != POOL_CONTINUE)
 			return status;
+
+		if (kind == 'Z' && frontend->no_forward && state == 1)
+		{
+			state = 0;
+		}
+
 	}
 	return POOL_CONTINUE;
 }
 
 static POOL_STATUS Query(POOL_CONNECTION *frontend, 
-						 POOL_CONNECTION_POOL *backend)
+						 POOL_CONNECTION_POOL *backend, char *query)
 {
-	char kind = 'Q';
 	char *string;
 	int len;
 	static char *sq = "show pool_status";
 
-	/* read actual query */
-	string = pool_read_string(frontend, &len, 0);
-	if (string == NULL)
-		return POOL_END;
+	if (query == NULL)
+	{
+		/* read actual query */
+		if (MAJOR(backend) == PROTO_MAJOR_V3)
+		{
+			if (pool_read(frontend, &len, sizeof(len)) < 0)
+				return POOL_END;
+			len = ntohl(len) - 4;
+			string = pool_read2(frontend, len);
+		}
+		else
+			string = pool_read_string(frontend, &len, 0);
+
+		if (string == NULL)
+			return POOL_END;
+	}
+	else
+	{
+		len = strlen(query)+1;
+		string = query;
+	}
+
+	pool_debug("Query: %s", string);
 
 	/* process status reporting? */
 	if (strncasecmp(sq, string, strlen(sq)) == 0)
 	{
 		pool_debug("process reporting");
-		process_reporting(frontend);
+		process_reporting(frontend, backend);
 		return POOL_CONTINUE;
 	}
 
+	/* load balance trick */
+	if (load_balance_enabled(backend, string))
+		start_load_balance(backend);
+
 	/* forward the query to the backend */
-	pool_write(MASTER(backend), &kind, 1);
+	pool_write(MASTER(backend), "Q", 1);
+
+	if (MAJOR(backend) == PROTO_MAJOR_V3)
+	{
+		int sendlen = htonl(len + 4);
+		pool_write(MASTER(backend), &sendlen, sizeof(sendlen));
+	}
+
 	if (pool_write_and_flush(MASTER(backend), string, len) < 0)
 	{
 		return POOL_END;
@@ -365,7 +480,13 @@ static POOL_STATUS Query(POOL_CONNECTION *frontend,
 			if (synchronize(MASTER(backend)))
 				return POOL_END;
 
-		pool_write(SECONDARY(backend), &kind, 1);
+		pool_write(SECONDARY(backend), "Q", 1);
+		if (MAJOR(backend) == PROTO_MAJOR_V3)
+		{
+			int sendlen = htonl(len + 4);
+			pool_write(SECONDARY(backend), &sendlen, sizeof(sendlen));
+		}
+
 		if (pool_write_and_flush(SECONDARY(backend), string, len) < 0)
 		{
 			return POOL_END;
@@ -377,17 +498,48 @@ static POOL_STATUS Query(POOL_CONNECTION *frontend,
 static POOL_STATUS ReadyForQuery(POOL_CONNECTION *frontend, 
 								 POOL_CONNECTION_POOL *backend, int send_ready)
 {
-	char kind;
+	int len;
+
+	pool_flush(frontend);
 
 	if (send_ready)
 	{
-		kind = 'Z';
-		if (pool_write_and_flush(frontend, &kind, 1) < 0)
+		pool_write(frontend, "Z", 1);
+
+		if (MAJOR(backend) == PROTO_MAJOR_V3)
 		{
-			return POOL_END;
+			int len;
+			char state;
+
+			if ((len = pool_read_message_length(backend)) < 0)
+				return POOL_END;
+
+			pool_debug("ReadyForQuery: message length: %d", len);
+
+			len = htonl(len);
+			pool_write(frontend, &len, sizeof(len));
+
+			state = pool_read_kind(backend);
+			if (state < 0)
+				return POOL_END;
+
+			/* set transaction state */
+			pool_debug("ReadyForQuery: transaction state: %c", state);
+			MASTER(backend)->tstate = state;
+			if (REPLICATION)
+				SECONDARY(backend)->tstate = state;
+
+			pool_write(frontend, &state, 1);
 		}
+
+		if (pool_flush(frontend))
+			return POOL_END;
 	}
 
+	/* end load balance mode */
+	if (in_load_balance)
+		end_load_balance(backend);
+
 	return ProcessFrontendResponse(frontend, backend);
 }
 
@@ -520,13 +672,6 @@ static int RowDescription(POOL_CONNECTION *frontend,
 		pool_write(frontend, &mod, sizeof(int));
 	}
 
-#ifdef NOT_USED
-	if (pool_flush(frontend) < 0)
-	{
-		return POOL_END;
-	}
-#endif
-
 	return num_fields;
 }
 
@@ -637,13 +782,6 @@ static POOL_STATUS AsciiRow(POOL_CONNECTION *frontend,
 		mask >>= 1;
 	}
 
-#ifdef NOT_USED
-	if (pool_flush(frontend) < 0)
-	{
-		return POOL_END;
-	}
-#endif
-
 	return POOL_CONTINUE;
 }
 
@@ -805,10 +943,17 @@ static POOL_STATUS CopyInResponse(POOL_CONNECTION *frontend,
 	POOL_STATUS status;
 
 	/* forward to the frontend */
-	if (pool_write_and_flush(frontend, "G", 1) < 0)
+	if (MAJOR(backend) == PROTO_MAJOR_V3)
 	{
-		return POOL_END;
+		if (SimpleForwardToFrontend('G', frontend, backend) != POOL_CONTINUE)
+			return POOL_END;
+		if (pool_flush(frontend) != POOL_CONTINUE)
+			return POOL_END;
 	}
+	else
+		if (pool_write_and_flush(frontend, "G", 1) < 0)
+			return POOL_END;
+
 	status = CopyDataRows(frontend, backend, 1);
 	return status;
 }
@@ -819,10 +964,17 @@ static POOL_STATUS CopyOutResponse(POOL_CONNECTION *frontend,
 	POOL_STATUS status;
 
 	/* forward to the frontend */
-	if (pool_write_and_flush(frontend, "H", 1) < 0)
+	if (MAJOR(backend) == PROTO_MAJOR_V3)
 	{
-		return POOL_END;
+		if (SimpleForwardToFrontend('H', frontend, backend) != POOL_CONTINUE)
+			return POOL_END;
+		if (pool_flush(frontend) != POOL_CONTINUE)
+			return POOL_END;
 	}
+	else
+		if (pool_write_and_flush(frontend, "H", 1) < 0)
+			return POOL_END;
+
 	status = CopyDataRows(frontend, backend, 0);
 	return status;
 }
@@ -841,12 +993,49 @@ static POOL_STATUS CopyDataRows(POOL_CONNECTION *frontend,
 	for (;;)
 	{
 		if (copyin)
-			string = pool_read_string(frontend, &len, 1);
+		{
+			if (MAJOR(backend) == PROTO_MAJOR_V3)
+			{
+				char kind;
+
+				if (pool_read(frontend, &kind, 1) < 0)
+					return POOL_END;
+				
+				SimpleForwardToBackend(kind, frontend, backend);
+
+				/* CopyData? */
+				if (kind == 'd')
+					continue;
+				else
+					break;
+			}
+			else
+				string = pool_read_string(frontend, &len, 1);
+		}
 		else
 		{
-			string = pool_read_string(MASTER(backend), &len, 1);
-			if (REPLICATION)
-				string = pool_read_string(SECONDARY(backend), &len, 1);
+			/* CopyOut */
+			if (MAJOR(backend) == PROTO_MAJOR_V3)
+			{
+				char kind;
+
+				if ((kind = pool_read_kind(backend)) < 0)
+					return POOL_END;
+				
+				SimpleForwardToFrontend(kind, frontend, backend);
+
+				/* CopyData? */
+				if (kind == 'd')
+					continue;
+				else
+					break;
+			}
+			else
+			{
+				string = pool_read_string(MASTER(backend), &len, 1);
+				if (REPLICATION)
+					string = pool_read_string(SECONDARY(backend), &len, 1);
+			}
 		}
 
 		if (string == NULL)
@@ -866,7 +1055,7 @@ static POOL_STATUS CopyDataRows(POOL_CONNECTION *frontend,
 		else
 			pool_write(frontend, string, len);			
 
-		if (len == 3)
+		if (len == PROTO_MAJOR_V3)
 		{
 			/* end of copy? */
 			if (string[0] == '\\' &&
@@ -1089,6 +1278,9 @@ static POOL_STATUS ProcessFrontendResponse(POOL_CONNECTION *frontend,
 	char fkind;
 	POOL_STATUS status;
 
+	if (frontend->no_forward != 0)
+		return POOL_CONTINUE;
+
 	if (pool_read(frontend, &fkind, 1) < 0)
 	{
 		pool_error("ProcessFrontendResponse: failed to read kind");
@@ -1100,20 +1292,35 @@ static POOL_STATUS ProcessFrontendResponse(POOL_CONNECTION *frontend,
 	switch (fkind)
 	{
 		case 'X':
+			if (MAJOR(backend) == PROTO_MAJOR_V3)
+			{
+				int len;
+				pool_read(frontend, &len, sizeof(len));
+			}
 			status = POOL_END;
 			break;
 
 		case 'Q':
-			status = Query(frontend, backend);
-			break;
-
-		case 'F':
-			status = FunctionCall(frontend, backend);
+			status = Query(frontend, backend, NULL);
 			break;
 
 		default:
-			pool_error("ProcessFrontendResponse: unknown message type %c(%02x)", fkind, fkind);
-			status = POOL_ERROR;
+			if (MAJOR(backend) == PROTO_MAJOR_V3)
+			{
+				status = SimpleForwardToBackend(fkind, frontend, backend);
+				if (pool_flush(MASTER(backend)))
+					status = POOL_ERROR;
+				if (REPLICATION)
+					if (pool_flush(SECONDARY(backend)))
+						status = POOL_ERROR;
+			}
+			else if (MAJOR(backend) == PROTO_MAJOR_V2 && fkind == 'F')
+				status = FunctionCall(frontend, backend);
+			else
+			{
+				pool_error("ProcessFrontendResponse: unknown message type %c(%02x)", fkind, fkind);
+				status = POOL_ERROR;
+			}
 			break;
 	}
 
@@ -1204,7 +1411,7 @@ int pool_check_fd(POOL_CONNECTION *cp, int notimeout)
 	return -1;
 }
 
-static void process_reporting(POOL_CONNECTION *frontend)
+static void process_reporting(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *backend)
 {
 	static char *cursorname = "blank";
 	static short num_fields = 3;
@@ -1215,6 +1422,8 @@ static void process_reporting(POOL_CONNECTION *frontend)
 	short n;
 	int i;
 	short s;
+	int len;
+	short colnum;
 
 	static char nullmap[2] = {0xff, 0xff};
 	int nbytes = (num_fields + 7)/8;
@@ -1231,7 +1440,7 @@ static void process_reporting(POOL_CONNECTION *frontend)
 
 	POOL_REPORT_STATUS status[MAXITEMS];
 
-	int nrows;
+	short nrows;
 	int size;
 	int hsize;
 
@@ -1329,67 +1538,166 @@ static void process_reporting(POOL_CONNECTION *frontend)
 
 	status[i].name = "replication_enabled";
 	snprintf(status[i].value, MAXVALLEN, "%d", pool_config.replication_enabled);
-	status[i].desc = "non 0 if operating in replication mode";
+	status[i].desc = "non 0 if actually operating in replication mode";
+	i++;
+
+	status[i].name = "load_balance_mode";
+	snprintf(status[i].value, MAXVALLEN, "%d", pool_config.load_balance_mode);
+	status[i].desc = "non 0 if operating in load balancing mode";
+	i++;
+
+	status[i].name = "replication_stop_on_mismatch";
+	snprintf(status[i].value, MAXVALLEN, "%d", pool_config.replication_stop_on_mismatch);
+	status[i].desc = "stop replication mode on fatal error";
 	i++;
 
 	nrows = i;
 
-	/* cursor response */
-	pool_write(frontend, "P", 1);
-	pool_write(frontend, cursorname, strlen(cursorname)+1);
+	if (MAJOR(backend) == PROTO_MAJOR_V2)
+	{
+		/* cursor response */
+		pool_write(frontend, "P", 1);
+		pool_write(frontend, cursorname, strlen(cursorname)+1);
+	}
 
 	/* row description */
 	pool_write(frontend, "T", 1);
+
+	if (MAJOR(backend) == PROTO_MAJOR_V3)
+	{
+		len = sizeof(num_fields) + sizeof(len);
+
+		for (i=0;i<num_fields;i++)
+		{
+			char *f = field_names[i];
+			len += strlen(f)+1;
+			len += sizeof(oid);
+			len += sizeof(colnum);
+			len += sizeof(oid);
+			len += sizeof(s);
+			len += sizeof(mod);
+			len += sizeof(s);
+		}
+
+		len = htonl(len);
+		pool_write(frontend, &len, sizeof(len));
+	}
+
 	n = htons(num_fields);
 	pool_write(frontend, &n, sizeof(short));
+
 	for (i=0;i<num_fields;i++)
 	{
 		char *f = field_names[i];
-		pool_write(frontend, f, strlen(f)+1);
 
-		pool_write(frontend, &oid, sizeof(oid));
+		pool_write(frontend, f, strlen(f)+1);		/* field name */
+
+		if (MAJOR(backend) == PROTO_MAJOR_V3)
+		{
+			pool_write(frontend, &oid, sizeof(oid));	/* table oid */
+			colnum = htons(i);
+			pool_write(frontend, &colnum, sizeof(colnum));	/* column number */
+		}
+
+		pool_write(frontend, &oid, sizeof(oid));		/* data type oid */
 		s = htons(fsize);
-		pool_write(frontend, &s, sizeof(fsize));
-		pool_write(frontend, &mod, sizeof(mod));
+		pool_write(frontend, &s, sizeof(fsize));		/* field size */
+		pool_write(frontend, &mod, sizeof(mod));		/* modifier */
+
+		if (MAJOR(backend) == PROTO_MAJOR_V3)
+		{
+			s = htons(0);
+			pool_write(frontend, &s, sizeof(fsize));	/* field format (text) */
+		}
 	}
 	pool_flush(frontend);
 
-	/* ascii row */
-	for (i=0;i<nrows;i++)
+	if (MAJOR(backend) == PROTO_MAJOR_V2)
 	{
-		pool_write(frontend, "D", 1);
-		pool_write_and_flush(frontend, nullmap, nbytes);
+		/* ascii row */
+		for (i=0;i<nrows;i++)
+		{
+			pool_write(frontend, "D", 1);
+			pool_write_and_flush(frontend, nullmap, nbytes);
 
-		size = strlen(status[i].name);
-		hsize = htonl(size+4);
-		pool_write(frontend, &hsize, sizeof(hsize));
-		pool_write(frontend, status[i].name, size);
+			size = strlen(status[i].name);
+			hsize = htonl(size+4);
+			pool_write(frontend, &hsize, sizeof(hsize));
+			pool_write(frontend, status[i].name, size);
 
-		size = strlen(status[i].value);
-		hsize = htonl(size+4);
-		pool_write(frontend, &hsize, sizeof(hsize));
-		pool_write(frontend, status[i].value, size);
+			size = strlen(status[i].value);
+			hsize = htonl(size+4);
+			pool_write(frontend, &hsize, sizeof(hsize));
+			pool_write(frontend, status[i].value, size);
 
-		size = strlen(status[i].desc);
-		hsize = htonl(size+4);
-		pool_write(frontend, &hsize, sizeof(hsize));
-		pool_write(frontend, status[i].desc, size);
+			size = strlen(status[i].desc);
+			hsize = htonl(size+4);
+			pool_write(frontend, &hsize, sizeof(hsize));
+			pool_write(frontend, status[i].desc, size);
+		}
+	}
+	else
+	{
+		/* data row */
+		for (i=0;i<nrows;i++)
+		{
+			pool_write(frontend, "D", 1);
+			len = sizeof(len) + sizeof(nrows);
+			len += sizeof(int) + strlen(status[i].name);
+			len += sizeof(int) + strlen(status[i].value);
+			len += sizeof(int) + strlen(status[i].desc);
+			len = htonl(len);
+			pool_write(frontend, &len, sizeof(len));
+			s = htons(3);
+			pool_write(frontend, &s, sizeof(s));
+
+			len = htonl(strlen(status[i].name));
+			pool_write(frontend, &len, sizeof(len));
+			pool_write(frontend, status[i].name, strlen(status[i].name));
+
+			len = htonl(strlen(status[i].value));
+			pool_write(frontend, &len, sizeof(len));
+			pool_write(frontend, status[i].value, strlen(status[i].value));
+			
+			len = htonl(strlen(status[i].desc));
+			pool_write(frontend, &len, sizeof(len));
+			pool_write(frontend, status[i].desc, strlen(status[i].desc));
+		}
 	}
 
 	/* complete command response */
 	pool_write(frontend, "C", 1);
+	if (MAJOR(backend) == PROTO_MAJOR_V3)
+	{
+		len = htonl(sizeof(len) + strlen("SELECT")+1);
+		pool_write(frontend, &len, sizeof(len));
+	}
 	pool_write(frontend, "SELECT", strlen("SELECT")+1);
 
 	/* ready for query */
 	pool_write(frontend, "Z", 1);
+	if (MAJOR(backend) == PROTO_MAJOR_V3)
+	{
+		len = htonl(sizeof(len) + 1);
+		pool_write(frontend, &len, sizeof(len));
+		pool_write(frontend, "I", 1);
+	}
 
 	pool_flush(frontend);
 }
 
 void pool_send_frontend_exits(POOL_CONNECTION_POOL *backend)
 {
+	int len;
+
 	pool_write(MASTER(backend), "X", 1);
 
+	if (MAJOR(backend) == PROTO_MAJOR_V3)
+	{
+		len = htonl(4);
+		pool_write(MASTER(backend), &len, sizeof(len));
+	}
+
 	/*
 	 * XXX we cannot call pool_flush() here since backend may already
 	 * close the socket and pool_flush() automatically invokes fail
@@ -1402,6 +1710,305 @@ void pool_send_frontend_exits(POOL_CONNECTION_POOL *backend)
 	if (REPLICATION)
 	{
 		pool_write(SECONDARY(backend), "X", 1);
+		if (MAJOR(backend) == PROTO_MAJOR_V3)
+		{
+			len = htonl(4);
+			pool_write(MASTER(backend), &len, sizeof(len));
+		}
 		fflush(SECONDARY(backend)->write_fd);
 	}
 }
+
+/*
+ * -------------------------------------------------------
+ * V3 functions
+ * -------------------------------------------------------
+ */
+POOL_STATUS SimpleForwardToFrontend(char kind, POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *backend)
+{
+	int len;
+	int sendlen;
+	char *p;
+
+	pool_write(frontend, &kind, 1);
+
+	len = pool_read_message_length(backend);
+	if (len < 0)
+	{
+		return POOL_END;
+	}
+
+	sendlen = htonl(len);
+	pool_write(frontend, &sendlen, sizeof(sendlen));
+
+	len -= 4;
+
+	p = pool_read2(MASTER(backend), len);
+	if (p == NULL)
+		return POOL_END;
+
+	if (REPLICATION)
+		if (pool_read2(SECONDARY(backend), len) == NULL)
+			return POOL_END;
+
+	return pool_write(frontend, p, len);
+}
+
+POOL_STATUS SimpleForwardToBackend(char kind, POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *backend)
+{
+	int len;
+	int sendlen;
+	char *p;
+
+	if (pool_write(MASTER(backend), &kind, 1))
+		return POOL_END;
+	if (REPLICATION)
+		if (pool_write(SECONDARY(backend), &kind, 1))
+			return POOL_END;
+
+	if (pool_read(frontend, &sendlen, sizeof(sendlen)))
+	{
+		return POOL_END;
+	}
+
+	len = ntohl(sendlen) - 4;
+
+	p = pool_read2(frontend, len);
+	if (p == NULL)
+		return POOL_END;
+
+	if (pool_write(MASTER(backend), &sendlen, sizeof(sendlen)))
+		return POOL_END;
+	if (pool_write(MASTER(backend), p, len))
+		return POOL_END;
+
+	if (REPLICATION)
+	{
+		if (pool_write(SECONDARY(backend), &sendlen, sizeof(sendlen)))
+			return POOL_END;
+		if (pool_write(SECONDARY(backend), p, len))
+			return POOL_END;
+	}
+
+	return POOL_CONTINUE;
+}
+
+POOL_STATUS ParameterStatus(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *backend)
+{
+	int len;
+	int sendlen;
+	char *p;
+	char *name;
+	char *value;
+
+	pool_write(frontend, "S", 1);
+
+	len = pool_read_message_length(backend);
+	if (len < 0)
+	{
+		return POOL_END;
+	}
+
+	sendlen = htonl(len);
+	pool_write(frontend, &sendlen, sizeof(sendlen));
+
+	len -= 4;
+
+	p = pool_read2(MASTER(backend), len);
+	if (p == NULL)
+		return POOL_END;
+
+	name = p;
+	value = p + strlen(name) + 1;
+
+	pool_debug("name: %s value: %s", name, value);
+
+	pool_add_param(&MASTER(backend)->params, name, value);
+
+#ifdef DEBUG
+	pool_param_debug_print(&MASTER(backend)->params);
+#endif
+
+	if (REPLICATION)
+		if (pool_read2(SECONDARY(backend), len) == NULL)
+			return POOL_END;
+
+	return pool_write(frontend, p, len);
+
+}
+
+/*
+ * reset backend status. return values are:
+ * 0: no query was issued 1: a query was issued 2: no more queries remain -1: error
+ */
+static int reset_backend(POOL_CONNECTION_POOL *backend, int qcnt)
+{
+#ifdef NO_RESET_ALL
+	static char *queries[] = {"ABORT"};
+#else
+	static char *queries[] = {"ABORT", "RESET ALL"};
+#endif
+
+	char *query;
+	int qn = sizeof(queries)/sizeof(char *);
+	int len;
+
+	if (qcnt >= qn)
+		return 2;
+
+	query = queries[qcnt];
+
+	/* if transaction state is idle, we don't need to issue ABORT */
+	if (TSTATE(backend) == 'I' && !strcmp("ABORT", query))
+		return 0;
+
+	if (Query(NULL, backend, query) != POOL_CONTINUE)
+		return -1;
+
+	return 1;
+}
+
+/*
+ * return non 0 if load balance is possible
+ */
+static int load_balance_enabled(POOL_CONNECTION_POOL *backend, char *sql)
+{
+	if (pool_config.load_balance_mode &&
+		REPLICATION &&
+		MAJOR(backend) == PROTO_MAJOR_V3 &&
+		TSTATE(backend) == 'I' &&
+		!strncasecmp(sql, "SELECT", 6))
+		return 1;
+	return 0;
+}
+
+/*
+ * start load balance mode
+ */
+static void start_load_balance(POOL_CONNECTION_POOL *backend)
+{
+	int i;
+	int master;
+
+	/* save backend connection slots */
+	for (i=0;i<backend->num;i++)
+	{
+		slots[i] = backend->slots[i];
+	}
+
+	/* temporary turn off replication mode */
+	REPLICATION = 0;
+
+	/* choose a master in random manner */
+	master = random() % backend->num;
+	backend->slots[0] = slots[master];
+	pool_debug("start_load_balance: selected master is %d", master);
+
+	/* start load balancing */
+	in_load_balance = 1;
+}
+
+/*
+ * finish load balance mode
+ */
+static void end_load_balance(POOL_CONNECTION_POOL *backend)
+{
+	int i;
+
+	/* restore backend connection slots */
+	for (i=0;i<backend->num;i++)
+	{
+		backend->slots[i] = slots[i];
+	}
+
+	/* turn on replication mode */
+	REPLICATION = 1;
+
+	in_load_balance = 0;
+	pool_debug("end_load_balance: end load balance mode");
+}
+
+/*
+ * send error message to frontend
+ */
+void pool_send_error_message(POOL_CONNECTION *frontend, int protoMajor,
+							 char *code,
+							 char *message,
+							 char *detail,
+							 char *hint,
+							 char *file,
+							 int line)
+{
+#define MAXDATA	1024
+#define MAXMSGBUF 128
+	if (protoMajor == PROTO_MAJOR_V2)
+	{
+		pool_write(frontend, "E", 1);
+		pool_write_and_flush(frontend, message, strlen(message)+1);
+	}
+	else if (protoMajor == PROTO_MAJOR_V3)
+	{
+		char data[MAXDATA];
+		char msgbuf[MAXMSGBUF];
+		char *p;
+		int len;
+		int thislen;
+		int sendlen;
+
+		len = 0;
+
+		pool_write(frontend, "E", 1);
+
+		/* error level */
+		thislen = snprintf(msgbuf, MAXMSGBUF, "SERROR");
+		memcpy(data +len, msgbuf, thislen+1);
+		len += thislen + 1;
+
+		/* code */
+		thislen = snprintf(msgbuf, MAXMSGBUF, "C%s", code);
+		memcpy(data +len, msgbuf, thislen+1);
+		len += thislen + 1;
+
+		/* message */
+		thislen = snprintf(msgbuf, MAXMSGBUF, "M%s", message);
+		memcpy(data +len, msgbuf, thislen+1);
+		len += thislen + 1;
+
+		/* detail */
+		if (*detail != '\0')
+		{
+			thislen = snprintf(msgbuf, MAXMSGBUF, "D%s", detail);
+			memcpy(data +len, msgbuf, thislen+1);
+			len += thislen + 1;
+		}
+
+		/* hint */
+		if (*hint != '\0')
+		{
+			thislen = snprintf(msgbuf, MAXMSGBUF, "H%s", hint);
+			memcpy(data +len, msgbuf, thislen+1);
+			len += thislen + 1;
+		}
+
+		/* file */
+		thislen = snprintf(msgbuf, MAXMSGBUF, "F%s", file);
+		memcpy(data +len, msgbuf, thislen+1);
+		len += thislen + 1;
+
+		/* line */
+		thislen = snprintf(msgbuf, MAXMSGBUF, "L%d", line);
+		memcpy(data +len, msgbuf, thislen+1);
+		len += thislen + 1;
+
+		/* stop null */
+		len++;
+		*(data + len) = '\0';
+
+		sendlen = len;
+		len = htonl(len + 4);
+		pool_write(frontend, &len, sizeof(len));
+		pool_write_and_flush(frontend, data, sendlen);
+	}
+	else
+		pool_error("send_error_message: unknown protocol major %d", protoMajor);
+}
